/*-
 * %%----------------------------------------------------------------------------------------------
 * Solidify Framework - Solidify Documentation - FieldDescriptorBuilder.java
 * SPDX-License-Identifier: GPL-2.0-or-later
 * %----------------------------------------------------------------------------------------------%
 * Copyright (C) 2017 - 2022 University of Geneva
 * %----------------------------------------------------------------------------------------------%
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public
 * License along with this program.  If not, see
 * <http://www.gnu.org/licenses/gpl-2.0.html>.
 * ----------------------------------------------------------------------------------------------%%
 */

package ch.unige.solidify.test.helper;

import static org.springframework.restdocs.payload.PayloadDocumentation.fieldWithPath;

import java.util.List;

import org.apache.tomcat.util.buf.StringUtils;
import org.springframework.restdocs.payload.FieldDescriptor;
import org.springframework.restdocs.snippet.Attributes;

import ch.unige.solidify.test.model.TypeField;
import ch.unige.solidify.util.StringTool;

public class FieldDescriptorBuilder {

  public static final String CONSTRAINTS = "constraints";
  public static final String ENUM_VALUE_QUOTE = "`";
  public static final char ENUM_VALUE_SEPARATOR = ',';
  public static final String FORMAT_PREFIX = "_Format_:";
  public static final String MANDATORY_VALUE = "mandatoryValue";
  public static final String NO_CONSTRAINT = "";
  public static final String READ_ONLY_VALUE = "readOnlyValue";
  public static final String TEST_STRING_VALUE = "test";
  public static final String VALUE_PREFIX = "_Values_:";
  protected static final String NO_TEST_VALUE = "NO_TEST_VALUE";

  // Test value list
  protected static final String TEST_DATE_FORMAT_VALUE = "2030-07-10T15:36:50.323+0200";
  protected static final String TEST_DATE_SIMPLE_VALUE = "2030-07-10";
  protected static final String TEST_NUMBER_VALUE = "11";
  protected static final String TEST_TRUE_VALUE = "true";
  protected static final String TEST_URL_VALUE = "https://www.unige.ch/";

  private final AbstractResourceDoc resourceDoc;

  private String name;
  private String description;
  private boolean mandatory;
  private TypeField type = TypeField.STRING;
  private List<String> enumValues;
  private String customConstraints;
  private String testValue;
  private boolean readOnly = true;

  public FieldDescriptorBuilder(AbstractResourceDoc resourceDoc) {
    this.resourceDoc = resourceDoc;
  }

  public FieldDescriptorBuilder name(String name) {
    this.name = name;
    return this;
  }

  public FieldDescriptorBuilder description(String description) {
    this.description = description;
    return this;
  }

  public FieldDescriptorBuilder mandatory(boolean mandatory) {
    this.mandatory = mandatory;
    return this;
  }

  public FieldDescriptorBuilder type(TypeField type) {
    this.type = type;
    return this;
  }

  public FieldDescriptorBuilder enumValues(List<String> enumValues) {
    this.enumValues = enumValues;
    return this;
  }

  public FieldDescriptorBuilder customConstraints(String customConstraints) {
    this.customConstraints = customConstraints;
    return this;
  }

  public FieldDescriptorBuilder testValue(String testValue) {
    this.testValue = testValue;
    return this;
  }

  public FieldDescriptorBuilder readOnly(boolean readOnly) {
    this.readOnly = readOnly;
    return this;
  }

  public FieldDescriptor build() {
    final String constraints = this.determineConstraint(customConstraints, type, enumValues);
    testValue = this.determineTestValue(testValue, type, enumValues);
    final FieldDescriptor fieldDescriptor = fieldWithPath(name).description(description);
    this.setFieldDescriptorAttributes(fieldDescriptor, mandatory, testValue, constraints, readOnly);
    resourceDoc.addToOpenApiExtraDocumentation(name, enumValues, type);
    return fieldDescriptor;
  }

  private String determineConstraint(String constraints, TypeField type, List<String> enumValues) {
    if (constraints != null) {
      return constraints;
    }

    if (enumValues != null) {
      return this.listValues(enumValues);
    }

    switch (type) {
      case DATE -> constraints = this.displayFormat(StringTool.DATE_FORMAT);
      case DATE_TIME -> constraints = this.displayFormat(StringTool.DATE_TIME_FORMAT);
      default -> constraints = NO_CONSTRAINT;
    }
    return constraints;
  }

  private String determineTestValue(String testValue, TypeField type, List<String> enumValues) {
    if (testValue != null) {
      if (testValue.equals(NO_TEST_VALUE)) {
        return null;
      }
      return testValue;
    }

    if (enumValues != null) {
      if (enumValues.isEmpty()) {
        return null;
      }
      return enumValues.get(0);
    }

    switch (type) {
      case INT32, INT64, FLOAT, DOUBLE -> testValue = TEST_NUMBER_VALUE;
      case STRING, BYTE, BINARY, PASSWORD -> testValue = TEST_STRING_VALUE;
      case BOOLEAN -> testValue = TEST_TRUE_VALUE;
      case DATE -> testValue = TEST_DATE_SIMPLE_VALUE;
      case DATE_TIME -> testValue = TEST_DATE_FORMAT_VALUE;
      case URL -> testValue = TEST_URL_VALUE;
      case ARRAY, UNDEFINED -> {}
    }
    return testValue;
  }

  private void setFieldDescriptorAttributes(FieldDescriptor fieldDescriptor, boolean mandatory,
          String testValue, String constraints, boolean readOnly) {
    fieldDescriptor.attributes(new Attributes.Attribute(CONSTRAINTS, constraints));
    // Add attribute only true because AsciiDoc test if the value is present
    if (mandatory) {
      fieldDescriptor.attributes(new Attributes.Attribute(MANDATORY_VALUE, String.valueOf(mandatory)));
    }
    fieldDescriptor.attributes(new Attributes.Attribute(READ_ONLY_VALUE, String.valueOf(readOnly)));

    this.setTestValue(fieldDescriptor, testValue);
  }

  private void setTestValue(FieldDescriptor fieldDescriptor, String testValue) {
    if (testValue == null) {
      return;
    }
    fieldDescriptor.attributes(new Attributes.Attribute(TEST_STRING_VALUE, testValue));
  }

  private String listValues(List<String> enumValues) {
    final List<String> enumValuesFormated = enumValues.stream()
            .map(val -> " " + ENUM_VALUE_QUOTE + val + ENUM_VALUE_QUOTE)
            .toList();
    return VALUE_PREFIX + StringUtils.join(enumValuesFormated, ENUM_VALUE_SEPARATOR);
  }

  protected String displayFormat(String format) {
    return FORMAT_PREFIX + " " + ENUM_VALUE_QUOTE + format + ENUM_VALUE_QUOTE;
  }

}
